from collections import defaultdict
import numpy as np
from tqdm import tqdm
import warnings
from .warnings import  NotIntendedUseWarning


class TreePathFinder():
    """Class to calculate the boundaries of a decision tree.

    It retrieves the structure from the decision tree

    """

    def __init__(self, estimator):
        self.estimator = estimator
        self.n_nodes = estimator.tree_.node_count
        self.children_left = estimator.tree_.children_left
        self.children_right = estimator.tree_.children_right
        self.feature = estimator.tree_.feature
        self.threshold = estimator.tree_.threshold
        self.is_leaves = self._find_leaves()
        self.decision_path = self.find_decision_to_leaves()
        self.bin_boundaries = self.find_bin_boundaries()


    def _find_leaves(self):
        # The tree structure can be traversed to compute various properties such
        # as the depth of each node and whether or not it is a leaf.
        n_nodes = self.n_nodes
        children_left = self.children_left
        children_right = self.children_right

        node_depth = np.zeros(shape=n_nodes, dtype=np.int64)
        is_leaves = np.zeros(shape=n_nodes, dtype=bool)
        stack = [(0, -1)]  # seed is the root node id and its parent depth
        while len(stack) > 0:
            node_id, parent_depth = stack.pop()
            node_depth[node_id] = parent_depth + 1

            # If we have a test node
            if (children_left[node_id] != children_right[node_id]):
                stack.append((children_left[node_id], parent_depth + 1))
                stack.append((children_right[node_id], parent_depth + 1))
            else:
                is_leaves[node_id] = True

        return is_leaves

    def find_parent(self, leaf_id):

        in_left = np.where(self.children_left == leaf_id)[0]
        in_right = np.where(self.children_right == leaf_id)[0]
        is_inleft = len(in_left) > 0
        is_inright = len(in_right) > 0

        if is_inleft & is_inright:
            raise ValueError(f"leaf with id {leaf_id} not found in tree")
        elif is_inleft:
            parent = in_left[0]
            operator = '<='
        elif is_inright:
            parent = in_right[0]
            operator = '>'
        else:
            parent = 0
            operator = 'None'

        threshold = self.threshold[parent]
        feature = self.feature[parent]

        return parent, threshold, operator, feature


    def find_decision_to_leaves(self):

        leaves_ids = np.where(self.is_leaves)[0]

        decision_path = defaultdict(list)
        for leaf_id in tqdm(leaves_ids):
            parent_id = -1
            node_id = leaf_id
            while parent_id != 0:
                path_step = self.find_parent(node_id)
                parent_id = path_step[0]
                decision_path[leaf_id].append(path_step)
                node_id = parent_id

        return decision_path

    def find_bin_boundaries(self):

        out_dict = dict()
        for leaf_id in self.decision_path.keys():
            one_leaf_decisions = self.decision_path[leaf_id]

            if '<=' not in [oper[2] for oper in one_leaf_decisions]:
                max_val = np.inf
            else:
                max_val = min([oper[1] for oper in one_leaf_decisions if oper[2] == '<='])

            if '>' not in [oper[2] for oper in one_leaf_decisions]:
                min_val = -np.inf
            else:
                min_val = max([oper[1] for oper in one_leaf_decisions if oper[2] == '>'])


            out_dict[leaf_id] = {
                'min': min_val,
                'max': max_val,
            }

        return out_dict


    def get_boundaries(self):

        # check how many features are there. There is always a unique negative value in the array of features
        # that corresponds to the index of the leaves.
        # Hence the total number of features in the tree is the length of the array -1
        n_features = len(np.unique(self.feature))-1

        if n_features>1:
            warning = f"This functionality is intended for trees fitted on 1 feature. The current tree is fitted " \
                f"with {n_features} features"
            warnings.warn(NotIntendedUseWarning(warning))

        return self.bin_boundaries


import itertools
import warnings

import pandas as pd
from tqdm import tqdm

from probatus.binning import SimpleBucketer, AgglomerativeBucketer, QuantileBucketer
from probatus.stat_tests import es, ks, psi, ad, sw
from probatus.utils.arrayfuncs import check_numeric_dtypes


class DistributionStatistics(object):
    """
    Wrapper that applies a statistical method and a binning strategy to data.

    Args:
    statistical_test (string): Statistical method to apply, statistical methods implemented:

        - 'ES': Epps-Singleton

        - 'KS': Kolmogorov-Smirnov statistic

        - 'PSI': Population Stability Index

        - 'SW': Shapiro-Wilk based difference statistic

        - 'AD': Anderson-Darling TS


    binning_strategy (Optional, string): Binning strategy to apply, binning strategies implemented:

        - 'simplebucketer': equally spaced bins

        - 'agglomerativebucketer': binning by applying the Scikit-learn implementation of Agglomerative Clustering

        - 'quantilebucketer': bins with equal number of elements

        - 'default': applies a default binning for a given stats_test. For all tests appart from PSI, no binning (None)
        is used. For PSI by default quantilebucketer is used.

        - None: no binning is applied. The test is computed based on original distribution.

    bin_count (Optional, int) In case binning_strategy is not None, specify the number of bins to be used by the binning
    strategy. By default 10 bins are used.

    Example:
     >>> d1 = np.histogram(np.random.normal(size=1000), 10)[0]
     >>> d2 = np.histogram(np.random.normal(size=1000), 10)[0]
     >>> myTest = DistributionStatistics('KS', 'simplebucketer', bin_count=10)
     >>> myTest.compute(d1, d2, verbose=True)
    """

    binning_strategy_dict = {
        "simplebucketer": SimpleBucketer,
        "agglomerativebucketer": AgglomerativeBucketer,
        "quantilebucketer": QuantileBucketer,
        None: None,
    }
    statistical_test_dict = {
        "ES": {
            "func": es,
            "name": "Epps-Singleton",
            "default_binning": None,
        },
        "KS": {
            "func": ks,
            "name": "Kolmogorov-Smirnov",
            "default_binning": None,
        },
        "AD": {
            "func": ad,
            "name": "Anderson-Darling TS",
            "default_binning": None,
        },
        "SW": {
            "func": sw,
            "name": "Shapiro-Wilk based difference",
            "default_binning": None,
        },
        "PSI": {
            "func": psi,
            "name": "Population Stability Index",
            "default_binning": "quantilebucketer",
        },
    }

    def __init__(self, statistical_test, binning_strategy="default", bin_count=10):
        self.statistical_test = statistical_test.upper()
        self.binning_strategy = binning_strategy
        self.bin_count = bin_count
        self.fitted = False

        # Initialize the statistical test
        if self.statistical_test not in self.statistical_test_dict:
            raise NotImplementedError(
                "The statistical test should be one of {}".format(
                    self.statistical_test_dict.keys()
                )
            )
        else:
            self.statistical_test_name = self.statistical_test_dict[
                self.statistical_test
            ]["name"]
            self._statistical_test_function = self.statistical_test_dict[
                self.statistical_test
            ]["func"]

        # Initialize the binning strategy
        if self.binning_strategy:
            self.binning_strategy = self.binning_strategy.lower()
            if self.binning_strategy == "default":
                self.binning_strategy = self.statistical_test_dict[
                    self.statistical_test
                ]["default_binning"]
            if self.binning_strategy not in self.binning_strategy_dict:
                raise NotImplementedError(
                    "The binning strategy should be one of {}".format(
                        list(self.binning_strategy_dict.keys())
                    )
                )
            else:
                binner = self.binning_strategy_dict[self.binning_strategy]
                if binner is not None:
                    self.binner = binner(bin_count=self.bin_count)

    def __repr__(self):
        repr_ = "DistributionStatistics object\n\tstatistical_test: {}".format(
            self.statistical_test
        )
        if self.binning_strategy:
            repr_ += "\n\tbinning_strategy: {}\n\tbin_count: {}".format(
                self.binning_strategy, self.bin_count
            )
        else:
            repr_ += "\n\tNo binning applied"
        if self.fitted:
            repr_ += "\nResults\n\tvalue {}-statistic: {}".format(
                self.statistical_test, self.statistic
            )
        if hasattr(self, "p_value"):
            repr_ += "\n\tp-value: {}".format(self.p_value)
        return repr_

    def compute(self, d1, d2, verbose=False, **kwargs):
        """
        Apply the statistical test and compute statistic value and p-value

        Args:
            d1: (np.array or pd.DataFrame): distribution 1.

            d2: (np.array or pd.DataFrame): distribution 2.

            verbose: (bool): Flag indicating whether prints should be shown.


        Returns:
            (Tuple of floats): statistic value and p_value. For PSI test the return is only statistic
        """
        check_numeric_dtypes(d1)
        check_numeric_dtypes(d2)

        # Bin the data
        if self.binning_strategy:
            self.binner.fit(d1)
            d1_preprocessed = self.binner.counts
            d2_preprocessed = self.binner.compute(d2)
        else:
            d1_preprocessed, d2_preprocessed = d1, d2

        # Perform the statistical test
        res = self._statistical_test_function(
            d1_preprocessed, d2_preprocessed, verbose=verbose, **kwargs
        )
        self.fitted = True

        # Check form of results and return
        if type(res) == tuple:
            self.statistic, self.p_value = res
            return self.statistic, self.p_value
        else:
            self.statistic = res
            return self.statistic


class AutoDist(object):
    """
    Class to automatically apply all implemented statistical distribution tests and binning strategies to (a
    selection of) features in two dataframes.

    Args:
        statistical_tests (Optional, str): Statistical tests to apply, either list of tests names, or 'all'. Statistical
        methods implemented:

            - 'ES': Epps-Singleton,

            - 'KS': Kolmogorov-Smirnov statistic,

            - 'PSI': Population Stability Index,

            - 'AD': Anderson-Darling TS.

        binning_strategies (Optional, str): Binning strategies to apply for each test, either list of tests names, 'all'
        or 'default'.
    
            Binning strategies that can be chosen:

                    - 'SimpleBucketer': equally spaced bins,

                    - 'AgglomerativeBucketer': binning by applying the Scikit-learn implementation of Agglomerative Clustering,

                    - 'QuantileBucketer': bins with equal number of elements,

                    - None: no binning is applied. Note that not all statistical tests will be performed since 
                      some () require binning strategies.
                    
                    - 'default': applies a default binning for a given stats_test. For all tests appart from PSI, no binning (None)
                    is used. For PSI by default quantilebucketer is used.

                    - 'all': each binning strategy is used for each statistical test

        bin_count (integer, None or list of integers): bin_count value(s) to be used, note that None can only be used when
        no bucketing strategy is applied.

    Example:
     >>> df1 = pd.DataFrame(np.random.normal(size=(1000, 2)), columns=['feat_0', 'feat_1'])
     >>> df2 = pd.DataFrame(np.random.normal(size=(1000, 2)), columns=['feat_0', 'feat_1'])
     >>> myAutoDist = AutoDist(statistical_tests='all', binning_strategies='all', bin_count=[10, 20])
     >>> res = myAutoDist.fit(df1, df2, columns=df1.columns)
    """

    def __init__(
        self, statistical_tests="all", binning_strategies="default", bin_count=10
    ):
        self.fitted = False

        # Initialize statistical tests to be performed
        if statistical_tests == "all":
            self.statistical_tests = list(
                DistributionStatistics.statistical_test_dict.keys()
            )
        elif isinstance(statistical_tests, str):
            self.statistical_tests = [statistical_tests]
        else:
            self.statistical_tests = statistical_tests

        # Initialize binning strategies to be used
        if binning_strategies == "all":
            self.binning_strategies = list(
                DistributionStatistics.binning_strategy_dict.keys()
            )
        elif isinstance(binning_strategies, str):
            self.binning_strategies = [binning_strategies]
        elif binning_strategies is None:
            self.binning_strategies = [None]
        else:
            self.binning_strategies = binning_strategies
        if not isinstance(bin_count, list):
            self.bin_count = [bin_count]
        else:
            self.bin_count = bin_count

    def __repr__(self):
        repr_ = "AutoDist object"
        if not self.fitted:
            repr_ += "\n\tAutoDist not fitted"
        if self.fitted:
            repr_ += "\n\tAutoDist fitted"
        repr_ += "\n\tstatistical_tests: {}".format(self.statistical_tests)
        repr_ += "\n\tbinning_strategies: {}".format(self.binning_strategies)
        repr_ += "\n\tbin_count: {}".format(self.bin_count)
        return repr_

    def compute(
        self,
        df1,
        df2,
        column_names=None,
        return_failed_tests=True,
        suppress_warnings=False,
    ):
        """
        Fit the AutoDist object to data; i.e. apply the statistical tests and binning strategies

        Args:

            df1 (pd.DataFrame): dataframe 1 for distribution comparison with dataframe 2.

            df2 (pd.DataFrame): dataframe 2 for distribution comparison with dataframe 1.

            column_names (list of str): list of columns in df1 and df2 that should be compared. If None, 
            all column names will be compared

            return_failed_tests (bool): remove tests in result that did not succeed.

            suppress_warnings (bool): whether to suppress warnings during the fit process.

        Returns:
            (pd.DataFrame): dataframe with results of the performed statistical tests and binning strategies.

        """
        if column_names is None:
            column_names = df1.columns.to_list()
            if len(set(column_names) - set(df2.columns)):
                raise Exception(
                    "column_names was set to None but columns in provided dataframes are different"
                )
        # Check if all columns in column_names are in df1 and df2
        elif len(set(column_names) - set(df1.columns)) or len(
            set(column_names) - set(df2.columns)
        ):
            raise Exception(
                "Not all columns in `column_names` are in the provided dataframes"
            )

        # Calculate statistics and p-values for all combinations
        result_all = pd.DataFrame()
        for col, stat_test, bin_strat, bins in tqdm(
            list(
                itertools.product(
                    column_names,
                    self.statistical_tests,
                    self.binning_strategies,
                    self.bin_count,
                )
            )
        ):
            if self.binning_strategies == ["default"]:
                bin_strat = DistributionStatistics.statistical_test_dict[stat_test][
                    "default_binning"
                ]
            dist = DistributionStatistics(
                statistical_test=stat_test, binning_strategy=bin_strat, bin_count=bins
            )
            try:
                if suppress_warnings:
                    warnings.filterwarnings("ignore")
                _ = dist.compute(df1[col], df2[col])
                if suppress_warnings:
                    warnings.filterwarnings("default")
                statistic = dist.statistic
                p_value = dist.p_value
            except:
                statistic, p_value = "an error occurred", None
                pass

            # Append result to results list
            result_ = {
                "column": col,
                "statistical_test": stat_test,
                "binning_strategy": bin_strat,
                "bin_count": bins,
                "statistic": statistic,
                "p_value": p_value,
            }
            result_all = result_all.append(result_, ignore_index=True)

        if not return_failed_tests:
            result_all = result_all[result_all["statistic"] != "an error occurred"]
        self.fitted = True
        self._result = result_all[
            [
                "column",
                "statistical_test",
                "binning_strategy",
                "bin_count",
                "statistic",
                "p_value",
            ]
        ]
        self._result["bin_count"] = self._result["bin_count"].astype(int)
        self._result.loc[
            self._result["binning_strategy"].isnull(), "bin_count"
        ] = 0
        self._result.loc[
            self._result["binning_strategy"].isnull(), "binning_strategy"
        ] = "no_bucketing"

        # Remove duplicates that appear if multiple bin numbers are passed, and binning strategy None

        self._result = self._result.\
            drop_duplicates(subset=['column', 'statistical_test', 'binning_strategy', 'bin_count'], keep='first')

        # create pivot table as final output
        self.result = pd.pivot_table(
            self._result,
            values=["statistic", "p_value"],
            index="column",
            columns=["statistical_test", "binning_strategy", "bin_count"],
            aggfunc="sum",
        )

        # flatten multi-index
        self.result.columns = [
            "_".join([str(x) for x in line]) for line in self.result.columns.values
        ]
        self.result.reset_index(inplace=True)
        return self.result



from probatus.metric_volatility import BaseVolatilityEstimator, TrainTestVolatility, SplitSeedVolatility,\
    BootstrappedVolatility, get_metric, sample_data, check_sampling_input
import pytest
import numpy as np
import pandas as pd
from unittest.mock import patch
import matplotlib.pyplot as plt
from probatus.stat_tests.distribution_statistics import DistributionStatistics
from probatus.utils import Scorer, NotFittedError
import sklearn

@pytest.fixture(scope='function')
def X_array():
    return np.array([[2, 1], [3, 2], [4, 3], [1, 2], [1, 1]])


@pytest.fixture(scope='function')
def y_list():
    return [1, 0, 0, 1, 1]


@pytest.fixture(scope='function')
def y_array(y_list):
    return np.array(y_list)


@pytest.fixture(scope='function')
def X_df(X_array):
    return pd.DataFrame(X_array, columns=['c1', 'c2'])


@pytest.fixture(scope='function')
def y_series(y_list):
    return pd.DataFrame(y_list)

@pytest.fixture(scope='function')
def iteration_results():
    iterations_cols = ['metric_name', 'train_score', 'test_score', 'delta_score']
    return pd.DataFrame([['roc_auc', 0.8, 0.7, 0.1], ['roc_auc', 0.7, 0.6, 0.1], ['roc_auc', 0.9, 0.8, 0.1],
                         ['accuracy', 1, 0.9, 0.1], ['accuracy', 0.8, 0.7, 0.1], ['accuracy', 0.9, 0.8, 0.1]],
                        columns=iterations_cols)

@pytest.fixture(scope='function')
def report():
    report_cols = ['train_mean', 'train_std', 'test_mean', 'test_std', 'delta_mean', 'delta_std']
    report_index = ['roc_auc', 'accuracy']
    return pd.DataFrame([[0.8, 0.08164, 0.7, 0.08164, 0.1, 0],
                         [0.9, 0.08164, 0.8, 0.08164, 0.1, 0]], columns=report_cols, index=report_index).astype(float)

@pytest.fixture(scope='function')
def iterations_train():
    return pd.Series([0.8, 0.7, 0.9], name='train_score')

@pytest.fixture(scope='function')
def iterations_test():
    return pd.Series([0.7, 0.6, 0.8], name='test_score')

@pytest.fixture(scope='function')
def iterations_delta():
    return pd.Series([0.1, 0.1, 0.1], name='delta_score')

def test_inits(mock_model):
    vol1 = SplitSeedVolatility(mock_model, metrics=['accuracy', 'roc_auc'], test_prc=0.3, n_jobs=2,
                              stats_tests_to_apply=['ES', 'KS'], random_state=1, iterations=20)

    assert id(vol1.model) == id(mock_model)
    assert vol1.test_prc == 0.3
    assert vol1.n_jobs == 2
    assert vol1.stats_tests_to_apply == ['ES', 'KS']
    assert vol1.random_state == 1
    assert vol1.iterations == 20
    assert len(vol1.stats_tests_objects) == 2
    assert len(vol1.scorers) == 2
    assert vol1.sample_train_test_split_seed is True

    vol2 = BootstrappedVolatility(mock_model, metrics='roc_auc', stats_tests_to_apply='KS', test_sampling_fraction=0.8)

    assert id(vol2.model) == id(mock_model)
    assert vol2.stats_tests_to_apply == ['KS']
    assert len(vol2.stats_tests_objects) == 1
    assert len(vol2.scorers) == 1
    assert vol2.sample_train_test_split_seed is False
    assert vol2.test_sampling_fraction == 0.8
    assert vol2.fitted is False
    assert vol2.iterations_results is None
    assert vol2.report is None

def test_base_fit(mock_model, X_df, y_series):
    vol = BaseVolatilityEstimator(mock_model, random_state=1)

    with patch('numpy.random.seed') as mock_seed:
        vol.fit(X_df, y_series)
        mock_seed.assert_called_with(1)

    assert vol.iterations_results is None
    assert vol.report is None
    assert vol.fitted is True


def test_compute(report, mock_model, iterations_train, iterations_test, iterations_delta):
    vol = BaseVolatilityEstimator(mock_model)

    with pytest.raises(NotFittedError):
        vol.compute()

    vol.fit()
    with pytest.raises(ValueError):
        vol.compute()

    vol.report = report

    pd.testing.assert_frame_equal(vol.compute(), report)
    pd.testing.assert_frame_equal(vol.compute(metrics=['roc_auc']), report.loc[['roc_auc']])
    pd.testing.assert_frame_equal(vol.compute(metrics='roc_auc'), report.loc[['roc_auc']])


def test_plot(report, mock_model, iterations_train, iterations_test, iterations_delta):
    num_figures_before = plt.gcf().number

    with patch.object(BaseVolatilityEstimator, 'compute', return_value=report.loc[['roc_auc']]) as mock_compute:
        with patch.object(BaseVolatilityEstimator, 'get_samples_to_plot',
                          return_value=(iterations_train, iterations_test, iterations_delta)) as mock_get_samples:

            vol = BaseVolatilityEstimator(mock_model)
            vol.fitted = True

            vol.plot(metrics='roc_auc')
            mock_compute.assert_called_with(metrics='roc_auc')
            mock_get_samples.assert_called_with(metric_name='roc_auc')


    num_figures_after = plt.gcf().number
    assert num_figures_after == num_figures_before + 1


def test_get_samples_to_plot(mock_model, iteration_results, iterations_train, iterations_test, iterations_delta):
    vol = BaseVolatilityEstimator(mock_model)
    vol.fitted = True
    vol.iterations_results=iteration_results

    train, test, delta = vol.get_samples_to_plot(metric_name='roc_auc')
    pd.testing.assert_series_equal(train, iterations_train)
    pd.testing.assert_series_equal(test, iterations_test)
    pd.testing.assert_series_equal(delta, iterations_delta)


def test_create_report(mock_model, iteration_results, report):
    vol = BaseVolatilityEstimator(mock_model)
    vol.fitted = True
    vol.iterations_results = iteration_results

    vol.create_report()
    pd.testing.assert_frame_equal(vol.report, report, check_less_precise=3)


def test_compute_mean_std_from_runs(mock_model, iteration_results):
    vol = BaseVolatilityEstimator(mock_model)
    results = vol.compute_mean_std_from_runs(iteration_results[iteration_results['metric_name'] == 'roc_auc'])
    expected_results = [0.8, 0.08164, 0.7, 0.08164, 0.1, 0]
    for idx, item in enumerate(results):
        assert pytest.approx(item, 0.01) == expected_results[idx]


def test_compute_stats_tests_values(mock_model, iteration_results):
    vol = BaseVolatilityEstimator(mock_model, stats_tests_to_apply=['KS'])

    with patch.object(DistributionStatistics, 'compute', return_value=(0.1, 0.05)):
        stats = vol.compute_stats_tests_values(iteration_results)

    assert stats[0] == 0.1
    assert stats[1] == 0.05


def test_fit_compute(mock_model, report, X_df, y_series):
    vol = BaseVolatilityEstimator(mock_model)

    with patch.object(BaseVolatilityEstimator, 'fit') as mock_fit:
        with patch.object(BaseVolatilityEstimator, 'compute', return_value=report) as mock_compute:
            result = vol.fit_compute(X_df, y_series)

            mock_fit.assert_called_with(X_df, y_series)
            mock_compute.assert_called_with()

    pd.testing.assert_frame_equal(result, report)


def test_fit_train_test_sample_seed(mock_model, X_df, y_series, X_array, y_array, iteration_results):
    vol = TrainTestVolatility(mock_model, metrics='roc_auc', iterations=3, sample_train_test_split_seed=True)

    with patch.object(BaseVolatilityEstimator, 'fit') as mock_base_fit:
        with patch.object(TrainTestVolatility, 'create_report') as mock_create_report:
            with patch('probatus.metric_volatility.volatility.assure_numpy_array', side_effect=[X_array, y_array]):
                with patch('probatus.metric_volatility.volatility.get_metric', side_effect=[iteration_results.iloc[[0]], iteration_results.iloc[[1]], iteration_results.iloc[[2]]]):

                    vol.fit(X_df, y_series)

                    mock_base_fit.assert_called_once()
                    mock_create_report.assert_called_once()

    pd.testing.assert_frame_equal(vol.iterations_results, iteration_results.iloc[[0, 1, 2]])


def test_get_metric(mock_model, X_df, y_series, X_array, y_array):
    split_seed = 1
    test_prc = 0.6
    with patch('probatus.metric_volatility.metric.assure_numpy_array', side_effect=[X_array, y_array]) \
            as mock_assure_array:
        with patch('probatus.metric_volatility.metric.train_test_split',
                   return_value=(X_array[[0, 1, 2]], X_array[[3, 4]], y_array[[0, 1, 2]], y_array[[3, 4]])) \
                as mock_split:
            with patch('probatus.metric_volatility.metric.sample_data',
                       side_effect=[(X_array[[0, 1, 1]], y_array[[0, 1, 1]]), (X_array[[3, 3]], y_array[[3, 3]])]) \
                    as mock_sample:
                with patch.object(Scorer, 'score', side_effect=[0.8, 0.7]):
                    output = get_metric(X_df, y_series, mock_model, test_size=test_prc, split_seed=split_seed,
                                        scorers=[Scorer('roc_auc')], train_sampling_type='bootstrap',
                                        test_sampling_type='bootstrap', train_sampling_fraction=1,
                                        test_sampling_fraction=1)
                    mock_assure_array.assert_called()
                    mock_split.assert_called_once()
                    mock_sample.assert_called()
                    mock_model.fit.assert_called()

    expected_output = pd.DataFrame([['roc_auc', 0.8, 0.7, 0.1]],
                                   columns=['metric_name', 'train_score', 'test_score', 'delta_score'])
    pd.testing.assert_frame_equal(expected_output, output)


def test_sample_data_no_sampling(X_array, y_array):
    with patch('probatus.metric_volatility.utils.check_sampling_input') as mock_sampling_input:
        X_out, y_out = sample_data(X_array, y_array, sampling_type=None, sampling_fraction=1)
        mock_sampling_input.assert_called_once()
    np.testing.assert_array_equal(X_out, X_array)
    np.testing.assert_array_equal(y_out, y_array)


def test_sample_data_bootstrap(X_array, y_array):
    with patch('probatus.metric_volatility.utils.check_sampling_input') as mock_sampling_input:
        X_out, y_out = sample_data(X_array, y_array, sampling_type='bootstrap', sampling_fraction=0.8)
        mock_sampling_input.assert_called_once()
    assert X_out.shape == (4, 2)
    assert y_out.shape == (4, )


def test_sample_data_sample(X_array, y_array):
    with patch('probatus.metric_volatility.utils.check_sampling_input') as mock_sampling_input:
        X_out, y_out = sample_data(X_array, y_array, sampling_type='subsample', sampling_fraction=1)
        mock_sampling_input.assert_called_once()
    np.testing.assert_array_equal(X_out, X_array)
    np.testing.assert_array_equal(y_out, y_array)


def test_check_sampling_input(X_array, y_array):
    with pytest.raises(ValueError):
        check_sampling_input('bootstrap', 0, 'dataset')
    with pytest.raises(ValueError):
        check_sampling_input('subsample', 0, 'dataset')
    with pytest.raises(ValueError):
        check_sampling_input('subsample', 1, 'dataset')
    with pytest.raises(ValueError):
        check_sampling_input('subsample', 10, 'dataset')
    with pytest.raises(ValueError):
        check_sampling_input('wrong_name', 0.5, 'dataset')



import numpy as np


def sample_data(X, y, sampling_type, sampling_fraction, dataset_name='dataset'):
    check_sampling_input(sampling_type, sampling_fraction, dataset_name)

    if sampling_type is None:
        return X, y

    number_of_samples = np.ceil(sampling_fraction * X.shape[0]).astype(int)
    array_index = list(range(X.shape[0]))

    if sampling_type == 'bootstrap':
        rows_indexes = np.random.choice(array_index, number_of_samples, replace=True)
    else:
        if sampling_fraction == 1 or number_of_samples == X.shape[0]:
            return X,y
        else:
            rows_indexes = np.random.choice(array_index, number_of_samples, replace=True)
    return X[rows_indexes], y[rows_indexes]


def check_sampling_input(sampling_type, fraction, dataset_name):
    if sampling_type is not None:
        if sampling_type == 'bootstrap':
            if fraction <= 0:
                raise(ValueError(f'For bootstrapping {dataset_name} fraction needs to be above 0'))
        elif sampling_type == 'subsample':
            if  fraction <= 0 or fraction >= 1:
                raise(ValueError(f'For bootstrapping {dataset_name} fraction needs to be be above 0 and below 1'))
        else:
            raise(ValueError('This sampling method is not implemented'))


from scipy import stats
from ..utils import assure_numpy_array
from probatus.stat_tests.utils import verbose_p_vals

@verbose_p_vals
def ad(d1, d2, verbose=False):
    """
    Calculates the Anderson-Darling TS on 2 distributions.

    Can be used on continuous or discrete distributions. Any binning/bucketing of the distributions/samples should be
    done before passing them to this function.

    Anderson & Darling 1954

    Advantages:
    - Unlike the KS, the AD (like the ES) can be used on both continuous & discrete distributions.
    - Works well even when dist has fewer than 25 observations.
    - More powerful than KS, especially for differences in the tails of distributions.

    Args:
        d1 (np.array or pandas.core.series.Series): first sample

        d2 (np.array or pandas.core.series.Series): second sample

        verbose (bool): helpful interpretation msgs printed to stdout (default False)

    Returns:
        (float, float): AD test stat and p-value of rejecting the null hypothesis (that the two distributions are identical)
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    ad, critical_values, pvalue = stats.anderson_ksamp([d1, d2])

    return ad, pvalue



def class_name_from_object(obj):
    return obj.__class__.__name__

def assure_list_of_strings(variable, variable_name):
    if isinstance(variable, list):
        return variable
    elif isinstance(variable, str):
        return [variable]
    else:
        raise(ValueError('{} needs to be either a string or list of strings.').format(variable_name))


def assure_list_values_allowed(variable, variable_name, allowed_values):
    for value in variable:
        if value not in allowed_values:
            raise(ValueError('Value {} in variable {} is not allowed').format(value, variable_name))



from .resemblance_model import BaseResemblanceModel, PermutationImportanceResemblance, SHAPImportanceResemblance
__all__ = ['BaseResemblanceModel', 'PermutationImportanceResemblance', 'SHAPImportanceResemblance']



import pandas as pd
import random
from scipy import stats
from ..utils import assure_numpy_array


def sw(d1, d2, verbose=False):
    """
    Examines whether deviation from normality of two distributions are significantly different. By using Shapiro-Wilk test
    as the basis.

    Args:
        d1 (np.ndarray or pd.core.series.Series) : first sample.

        d2 (np.ndarray or pd.core.series.Series) : second sample.

        verbose (bool)                           : helpful interpretation msgs printed to stdout (default False).

    Returns:
        (float, float): SW test stat and p-value of rejecting the null hypothesis (that the two distributions are identical).
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    if len(d1) > 5000:
        d1 = pd.Series(random.choices(d1, k=5000))
    if len(d2) > 5000:
        d2 = pd.Series(random.choices(d2, k=5000))

    delta = stats.shapiro(d1)[0] - stats.shapiro(d2)[0]

    d1 = pd.Series(d1)
    d2 = pd.Series(d2)

    MOT = pd.concat([d1, d2])
    n1 = d1.shape[0]
    n2 = d2.shape[0]

    def ran_delta(n1, n2):
        take_ran = lambda n: random.sample(range(MOT.shape[0]), n)
        ran_1 = MOT.iloc[
            take_ran(n1),
        ]
        ran_2 = MOT.iloc[
            take_ran(n2),
        ]
        delta_ran = stats.shapiro(ran_1)[0] - stats.shapiro(ran_2)[0]
        return delta_ran

    collect = [ran_delta(n1, n2) for a in range(100)]
    collect = pd.Series(list(collect))
    delta_p_value = 1 - stats.percentileofscore(collect, delta) / 100

    quants = [0.025, 0.975]
    sig_vals = list(collect.quantile(quants))

    if verbose:

        if delta < sig_vals[0] or delta > sig_vals[1]:
            print(
                "\nShapiro_Difference | Null hypothesis : <delta is not different from zero> REJECTED."
            )
            print("\nDelta is outside 95% CI -> Distributions very different.")
        else:
            print(
                "\nShapiro_Difference | Null hypothesis : <delta is not different from zero> NOT REJECTED."
            )
            print("\nDelta is inside 95% CI -> Distributions are not different.")

    return delta, delta_p_value



class ApproximationWarning(Warning):
    def __init__(self, message):
        self.message = message


class NotIntendedUseWarning(Warning):
    def __init__(self, message):
        self.message = message


from .exceptions import NotFittedError, DimensionalityError, UnsupportedModelError
from .scoring import Scorer, get_scorers
from .arrayfuncs import assure_numpy_array, assure_pandas_df, check_1d, warn_if_missing, check_numeric_dtypes
from .warnings import ApproximationWarning
from ._utils import class_name_from_object, assure_list_of_strings, assure_list_values_allowed
from .tree import TreePathFinder
from .plots import plot_distributions_of_feature

__all__ = ['NotFittedError', 'DimensionalityError', 'UnsupportedModelError', 'Scorer', 'assure_numpy_array', 'assure_pandas_df', 'check_1d',
           'ApproximationWarning', 'class_name_from_object', 'get_scorers', 'assure_list_of_strings',
           'assure_list_values_allowed', 'warn_if_missing','TreePathFinder', 'check_numeric_dtypes',
           'plot_distributions_of_feature']



import numpy as np
import pandas as pd
import numbers

import pytest

from probatus.stat_tests import ks, psi, DistributionStatistics, AutoDist


def test_distribution_statistics_base():
    with pytest.raises(NotImplementedError):
        assert DistributionStatistics('doesnotexist', 'SimpleBucketer', bin_count=10)
    with pytest.raises(NotImplementedError):
        assert DistributionStatistics('psi', 'doesnotexist', bin_count=10)
    myTest = DistributionStatistics('psi', 'SimpleBucketer', bin_count=10)
    assert repr(myTest).startswith('DistributionStatistics')


def test_distribution_statistics_psi():
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.weibull(1, size=1000) - 1, 10)[0]
    myTest = DistributionStatistics('psi', 'SimpleBucketer', bin_count=10)
    assert not myTest.fitted
    psi_test, p_value_test = myTest.compute(d1, d2)
    assert myTest.fitted
    assert isinstance(psi_test, numbers.Number)


def test_distribution_statistics_tuple_output():
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.weibull(1, size=1000) - 1, 10)[0]
    myTest = DistributionStatistics('ks', 'SimpleBucketer', bin_count=10)
    assert not myTest.fitted
    res = myTest.compute(d1, d2)
    assert myTest.fitted
    assert isinstance(res, tuple)


def test_distribution_statistics_ks_no_binning():
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.weibull(1, size=1000) - 1, 10)[0]
    myTest = DistributionStatistics('ks', binning_strategy=None)
    assert not myTest.fitted
    res = myTest.compute(d1, d2)
    assert myTest.fitted
    assert isinstance(res, tuple)


def test_distribution_statistics_attributes_psi():
    a = np.random.normal(size=1000)
    b = np.random.normal(size=1000)
    d1 = np.histogram(a, 10)[0]
    d2 = np.histogram(b, 10)[0]
    myTest = DistributionStatistics('psi', binning_strategy=None)
    _ = myTest.compute(d1, d2, verbose=False)
    psi_value_test, p_value_test = psi(d1, d2, verbose=False)
    assert myTest.statistic == psi_value_test


def test_distribution_statistics_attributes_ks():
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.normal(size=1000), 10)[0]
    myTest = DistributionStatistics('ks', binning_strategy=None)
    _ = myTest.compute(d1, d2, verbose=False)
    ks_value, p_value = ks(d1, d2)
    assert myTest.statistic == ks_value


def test_distribution_statistics_autodist_base():
    nr_features = 2
    size = 1000
    np.random.seed(0)
    df1 = pd.DataFrame(np.random.normal(size=(size, nr_features)), columns=[f'feat_{x}' for x in range(nr_features)])
    df2 = pd.DataFrame(np.random.normal(size=(size, nr_features)), columns=[f'feat_{x}' for x in range(nr_features)])
    features = df1.columns
    myAutoDist = AutoDist(statistical_tests='all', binning_strategies='all', bin_count=[10, 20])
    assert repr(myAutoDist).startswith('AutoDist')
    assert not myAutoDist.fitted
    res = myAutoDist.compute(df1, df2, column_names=features)
    assert myAutoDist.fitted
    pd.testing.assert_frame_equal(res, myAutoDist.result)
    assert isinstance(res, pd.DataFrame)
    assert res['column'].values.tolist() == features.to_list()

    dist = DistributionStatistics(statistical_test='ks', binning_strategy='simplebucketer', bin_count=10)
    dist.compute(df1['feat_0'], df2['feat_0'])
    assert dist.p_value == res.loc[res['column'] == 'feat_0', 'p_value_KS_simplebucketer_10'][0]
    assert dist.statistic == res.loc[res['column'] == 'feat_0', 'statistic_KS_simplebucketer_10'][0]

    dist = DistributionStatistics(statistical_test='ks', binning_strategy=None, bin_count=10)
    dist.compute(df1['feat_0'], df2['feat_0'])
    assert dist.p_value == res.loc[res['column'] == 'feat_0', 'p_value_KS_no_bucketing_0'][0]
    assert dist.statistic == res.loc[res['column'] == 'feat_0', 'statistic_KS_no_bucketing_0'][0]


def test_distribution_statistics_autodist_column_names_error():
    df1 = pd.DataFrame({'feat_0': [1, 2, 3, 4, 5], 'feat_1': [5, 6, 7, 8, 9]})
    df2 = df1
    features = df1.columns.values.tolist() + ['missing_feature']
    myAutoDist = AutoDist()
    with pytest.raises(Exception):
        assert myAutoDist.compute(df1, df2, column_names=features)

    df1 = pd.DataFrame({'feat_0': [1, 2, 3, 4, 5], 'feat_1': [5, 6, 7, 8, 9]})
    df2 = df1.copy()
    df1['feat_2'] = 0
    features = df2.columns.values.tolist() + ['missing_feature']
    myAutoDist = AutoDist()
    with pytest.raises(Exception):
        assert myAutoDist.compute(df1, df2, column_names=features)


def test_distribution_statistics_autodist_return_failed_tests():
    df1 = pd.DataFrame({'feat_0': [1, 2, 3, 4, 5], 'feat_1': [5, 6, 7, 8, 9]})
    df2 = df1
    features = df1.columns.values.tolist()
    myAutoDist = AutoDist(binning_strategies="all")
    res = myAutoDist.compute(df1, df2, column_names=features, return_failed_tests=True)
    assert res.isin(['an error occurred']).any().any()
    res = myAutoDist.compute(df1, df2, column_names=features, return_failed_tests=False)
    assert not res.isin(['an error occurred']).any().any()
    
def test_distribution_statistics_autodist_default():
    df1 = pd.DataFrame({'feat_0': [1, 2, 3, 4, 5], 'feat_1': [5, 6, 7, 8, 9]})
    df2 = df1
    features = df1.columns.values.tolist()
    myAutoDist = AutoDist(binning_strategies="default", bin_count=10)
    res = myAutoDist.compute(df1, df2, column_names=features)
    for stat_test, stat_info in DistributionStatistics.statistical_test_dict.items():
        if stat_info['default_binning']:
            assert f"p_value_{stat_test}_{stat_info['default_binning']}_10" in res.columns
        else:
            assert f"p_value_{stat_test}_no_bucketing_0" in res.columns

    assert "p_value_agglomerativebucketer_10" not in res.columns
    assert res.shape == (len(df1.columns), 1 + 2 * len(DistributionStatistics.statistical_test_dict))

def test_distribution_statistics_autodist_init():
    myAutoDist = AutoDist(statistical_tests='all', binning_strategies='all')
    assert isinstance(myAutoDist.statistical_tests, list)
    myAutoDist = AutoDist(statistical_tests='ks', binning_strategies='all')
    assert myAutoDist.statistical_tests == ['ks']
    myAutoDist = AutoDist(statistical_tests=['ks', 'psi'], binning_strategies='all')
    assert myAutoDist.statistical_tests == ['ks', 'psi']

    myAutoDist = AutoDist(statistical_tests='all', binning_strategies='all')
    assert isinstance(myAutoDist.binning_strategies, list)
    myAutoDist = AutoDist(statistical_tests='all', binning_strategies='quantilebucketer')
    assert myAutoDist.binning_strategies == ['quantilebucketer']
    myAutoDist = AutoDist(statistical_tests='all', binning_strategies=['quantilebucketer', 'simplebucketer'])
    assert myAutoDist.binning_strategies == ['quantilebucketer', 'simplebucketer']



from .ad import ad
from .es import es
from .sw import sw
from .ks import ks
from .psi import psi
from .distribution_statistics import DistributionStatistics
from .distribution_statistics import AutoDist

__all__ = ["ks", "psi", "ad", "es", "sw", "DistributionStatistics", "AutoDist"]



import numpy as np
import pandas as pd
from probatus.stat_tests import ks, psi, es, ad, sw

from probatus.binning import binning



def test_psi_returns_zero():
    x = np.random.normal(size=1000)
    myBucketer = binning.QuantileBucketer(bin_count=10)
    myBucketer.fit(x)
    d1 = myBucketer.counts
    d2 = d1
    psi_test, p_value_test = psi(d1, d2, verbose=False)
    assert psi_test == 0.0


def test_psi_returns_large():
    d1 = np.histogram(np.random.normal(size=1000), 10)[0]
    d2 = np.histogram(np.random.weibull(1, size=1000) - 1, 10)[0]
    psi_test, p_value_test = psi(d1, d2, verbose=False)
    assert psi_test > 1.0


def test_ks_returns_one():
    d1 = np.random.normal(size=1000)
    d2 = d1
    assert ks(d1, d2)[1] == 1.0


def test_ks_accepts_pd_series():
    d1 = pd.Series(np.random.normal(size=1000))
    d2 = d1
    assert ks(d1, d2)[1] == 1.0


def test_ks_returns_small():
    d1 = np.random.normal(size=1000)
    d2 = np.random.weibull(1, size=1000) - 1
    assert ks(d1, d2)[1] < 0.001


def test_es_returns_one():
    d1 = np.random.normal(size=1000)
    d2 = d1
    assert es(d1, d2)[1] == 1.0


def test_es_returns_small():
    d1 = np.random.normal(size=1000)
    d2 = np.random.weibull(1, size=1000) - 1
    assert es(d1, d2)[1] < 0.001


def test_ad_returns_big():
    d1 = np.random.normal(size=1000)
    d2 = d1
    assert ad(d1, d2)[1] >= 0.25


def test_ad_returns_small():
    d1 = np.random.normal(size=1000)
    d2 = np.random.weibull(1, size=1000) - 1
    assert ad(d1, d2)[1] <= 0.001


def test_sw_returns_zero():
    d1 = np.random.normal(size=1000)
    d2 = d1
    assert sw(d1, d2)[0] == 0
    


import functools

def verbose_p_vals(func):
    """Decorator to enable verbose printing of p-vals"""

    @functools.wraps(func)
    def wrapper_verbose_p_vals(*args, **kwargs):
        test_name = func.__name__.upper()

        stat, pvalue = func(*args, **kwargs)

        if "verbose" in kwargs and kwargs["verbose"] is True:
            print("\n{}: pvalue =".format(test_name), pvalue)
            if pvalue < 0.01:
                print(
                    "\n{}: Null hypothesis rejected with 99% confidence. Distributions very different.".format(
                        test_name
                    )
                )
            elif pvalue < 0.05:
                print(
                    "\n{}: Null hypothesis rejected with 95% confidence. Distributions different.".format(
                        test_name
                    )
                )
            else:
                print(
                    "\n{}: Null hypothesis cannot be rejected. Distributions not statistically different.".format(
                        test_name
                    )
                )

        return stat, pvalue

    return wrapper_verbose_p_vals


from .metric import get_metric
from .volatility import BaseVolatilityEstimator, TrainTestVolatility, BootstrappedVolatility, SplitSeedVolatility
from .utils import sample_data, check_sampling_input


__all__ = ['get_metric', 'BaseVolatilityEstimator', 'TrainTestVolatility', 'BootstrappedVolatility',
           'SplitSeedVolatility', 'sample_data', 'check_sampling_input']



from scipy import stats
from ..utils import assure_numpy_array
from probatus.stat_tests.utils import verbose_p_vals

@verbose_p_vals
def ks(d1, d2, verbose=False):
    """
    Calculates the Kolmogorov-Smirnov statistic on 2 samples. Any binning/bucketing of the distributions/samples
    should be done before passing them to this function.

    Args:
        d1 (np.ndarray or pd.core.series.Series) : first sample.

        d2 (np.ndarray or pd.core.series.Series) : second sample.

        verbose (bool)                           : helpful interpretation msgs printed to stdout (default False).

    Returns:
        (float, float): KS test stat and p-value of rejecting the null hypothesis (that the two distributions are identical)
    """

    # Perform data checks
    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    # Perform statistical tests
    ks, pvalue = stats.ks_2samp(d1, d2)

    return ks, pvalue



class NotFittedError(Exception):
    def __init__(self, message):
        self.message = message


class DimensionalityError(Exception):
    def __init__(self, message):
        self.message = message


class UnsupportedModelError(Exception):
    def __init__(self, message):
        self.message = message


from sklearn.metrics import get_scorer


def get_scorers(metrics):
    """
    Returns Scorers list based on the input metrics list

    Args:
        metrics (string, list of strings, Scorer or list of Scorers): Metrics for which the scorers should be returned.
        It can be either a single metric name or list of names metric names and needs to be aligned with predefined
        classification scorers names in sklearn, see the `sklearn documentation <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
        In case a custom metric is used, one can create own Scorer (probatus.utils) and provide as a metric.

    Returns:
        list of Scorer objects: list of scorers that can be used for scoring models
    """
    scorers = []
    if isinstance(metrics, list):
        for metric in metrics:
            scorers = _append_single_metric_to_scorers(metric, scorers)
    else:
        scorers = _append_single_metric_to_scorers(metrics, scorers)
    return scorers


def _append_single_metric_to_scorers(metric, scorers):
    """
    Appends single metric to scorers list

    Args:
        metrics (string or Scorer object): metric that need to be added to scorers list.
        scorers (list of Scorer objects): list of scorers.

    Returns:
        list of Scorer objects: list of scorers that can be used for scoring models, extended by one element.
    """
    if isinstance(metric, str):
        scorers.append(Scorer(metric))
    elif isinstance(metric, Scorer):
        scorers.append(metric)
    else:
        raise (ValueError('The metrics should contain either strings or Scorer class'))
    return scorers


class Scorer:
    """
    Scores the samples model based on the provided metric name

    Args:
        metric_name (str):  Name of the metric used to evaluate the model. If the custom_scorer is not passed, the
        metric name needs to be aligned with predefined classification scorers names in sklearn, see the
        `sklearn documentation <https://scikit-learn.org/stable/modules/model_evaluation.html>`_

        custom_scorer (Optional sklearn.metrics Scorer callable): Object that can score samples.

    Examples:
        >>> from probatus.utils import Scorer
        >>> from sklearn.metrics import make_scorer
        >>> import numpy as np
        >>> scorer1 = Scorer('roc_auc')
        >>> def custom_metric(y_true, y_pred):
        >>>     return np.sum(y_true == y_pred)
        >>> scorer2 = Scorer('custom_metric', custom_scorer=make_scorer(custom_metric))
    """

    def __init__(self, metric_name, custom_scorer=None):
        self.metric_name = metric_name
        if custom_scorer is not None:
            self.scorer = custom_scorer
        else:
            self.scorer = get_scorer(self.metric_name)

    def score(self, model, X, y):
        """
        Scores the samples model based on the provided metric name

        Args:
            model (model object): Model to be scored.

            X (array-like of shape (n_samples,n_features)):  Samples on which the model is scored.

            y (array-like of shape (n_samples,)):  Labels on which the model is scored.

        Returns:
            float: Score returned by the model
        """
        return self.scorer(model, X, y)


from scipy import stats

from ..utils import assure_numpy_array
from probatus.stat_tests.utils import verbose_p_vals

@verbose_p_vals
def es(d1, d2, verbose=False):
    """
    Calculates the Epps-Singleton test statistic on 2 distributions. Can be used on continuous or discrete
    distributions. Any binning/bucketing of the distributions/samples should be done before passing them to this
    function.

    Whereas KS relies on the empirical distribution function, ES is based on the empirical characteristic function
    (Epps & Singleton 1986, Goerg & Kaiser 2009).

    Advantages:

    - Unlike the KS, the ES can be used on both continuous & discrete distributions.

    - ES has higher power (vs KS) in many examples.

    Disadvantages:
    - Not recommended for fewer than 25 observations. Instead, use the Anderson-Darling TS. (However, ES can still be
    used for small samples. A correction factor is applied so that the asymptotic TS distribution more closely follows
    the chi-squared distribution, such that p-values can be computed.)

    Args:
        d1 (np.array or pandas.core.series.Series) : first sample.

        d2 (np.array or pandas.core.series.Series) : second sample.

        verbose (bool) : helpful interpretation msgs printed to stdout (default False).

    Returns:
        (float, float): ES test stat and p-value of rejecting the null hypothesis (that the two distributions are identical)
    """

    d1 = assure_numpy_array(d1)
    d2 = assure_numpy_array(d2)

    es, pvalue = stats.epps_singleton_2samp(d1, d2)

    return es, pvalue



import pandas as pd
import numpy as np
from probatus.utils import assure_numpy_array
from probatus.metric_volatility.utils import sample_data
from sklearn.model_selection import train_test_split


def get_metric(X, y, model, test_size, split_seed, scorers, train_sampling_type=None, test_sampling_type=None,
               train_sampling_fraction=1, test_sampling_fraction=1):
    """
    Draws random train/test sample from the data using random seed and calculates metric of interest.

    Args:
        X: (np.array or pd.DataFrame) with features
        y: (np.array or pd.Series) with targets
        test_size: (float) fraction of data used for testing the model
        split_seed: (int) randomized seed used for splitting data
        scorers: (list of Scorers) list of Scorer objects used to score the trained model
        train_sampling_type: (Optional str) string indicating what type of sampling should be applied on train set:
                - None indicates that no additional sampling is done after splitting data
                - 'bootstrap' indicates that sampling with replacement will be performed on train data
                - 'subsample': indicates that sampling without repetition will be performed  on train data
        test_sampling_type: (Optional str) string indicating what type of sampling should be applied on test set:
                - None indicates that no additional sampling is done after splitting data
                - 'bootstrap' indicates that sampling with replacement will be performed on test data
                - 'subsample': indicates that sampling without repetition will be performed  on test data
        train_sampling_fraction: (Optional float): fraction of train data sampled, if sample_train_type is not None.
                Default value is 1
        test_sampling_fraction: (Optional float): fraction of test data sampled, if sample_test_type is not None.
                Default value is 1

    Returns: 
        (pd.Dataframe) Dataframe with results for a given model trained. Rows indicate the metric measured and columns
         ther esults

    """

    if not (isinstance(X, np.ndarray) or isinstance(X, pd.DataFrame)):
        X = assure_numpy_array(X)
    if not (isinstance(X, np.ndarray) or isinstance(X, pd.Series)):
        y = assure_numpy_array(y)

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=split_seed, stratify=y)

    # Sample data based on the input arguments
    X_train, y_train = sample_data(X=X_train, y=y_train, sampling_type=train_sampling_type,
                                   sampling_fraction=train_sampling_fraction, dataset_name='train')
    X_test, y_test = sample_data(X=X_test, y=y_test, sampling_type=test_sampling_type,
                                 sampling_fraction=test_sampling_fraction, dataset_name='test')

    model = model.fit(X_train, y_train)

    results_columns = ['metric_name', 'train_score', 'test_score', 'delta_score']
    results = pd.DataFrame([], columns=results_columns)

    for scorer in scorers:
        score_train = scorer.score(model, X_train, y_train)
        score_test = scorer.score(model, X_test, y_test)
        score_delta = score_train - score_test

        results = results.append(pd.DataFrame([[scorer.metric_name, score_train, score_test, score_delta]],
                                              columns=results_columns))
    return results




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from probatus.metric_volatility.metric import get_metric
from joblib import Parallel, delayed
from tqdm.auto import tqdm
from probatus.utils import assure_numpy_array, NotFittedError, get_scorers, assure_list_of_strings,\
    assure_list_values_allowed
from probatus.metric_volatility.utils import check_sampling_input
from probatus.stat_tests import DistributionStatistics
import warnings


class BaseVolatilityEstimator(object):
    """
    Base object for estimating volatility estimation. This class is a base class, therefore should cannot be used on its
    own.

    Args:
        model (model object): Binary classification model or pipeline.

        metrics (string, list of strings, Scorer or list of Scorers): Metrics for which the score is calculated.
        It can be either a name or list of names metric names and needs to be aligned with predefined classification
        scorers names in sklearn, see the `sklearn documentation <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
        In case a custom metric is used, one can create own Scorer (probatus.utils) and provide as a metric.
        By default 'roc_auc' is measured.

        test_prc (float, optional): Percentage of input data used as test. By default 0.25.

        n_jobs (int, optional): Number of parallel executions. If -1 use all available cores. By default 1.

        stats_tests_to_apply (None, string or list of strings, optional): List of tests to apply. Available options:

                    - 'ES': Epps-Singleton,

                    - 'KS': Kolmogorov-Smirnov statistic,

                    - 'PSI': Population Stability Index,

                    - 'SW': Shapiro-Wilk based difference statistic,

                    - 'AD': Anderson-Darling TS.

        random_state (int, optional): The seed used by the random number generator.
    """
    def __init__(self, model, metrics='roc_auc', test_prc=0.25, n_jobs=1, stats_tests_to_apply=None, random_state=42):
        self.model = model
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.test_prc = test_prc
        self.iterations_results = None
        self.report = None
        self.fitted = False
        self.allowed_stats_tests = list(DistributionStatistics.statistical_test_dict.keys())

        # TODO set reasonable default value for the parameter, to choose the statistical test for the user for different
        #  ways to compute volatility
        if stats_tests_to_apply is not None:
            self.stats_tests_to_apply = assure_list_of_strings(stats_tests_to_apply, 'stats_tests_to_apply')
            assure_list_values_allowed(variable=self.stats_tests_to_apply,
                                       variable_name='stats_tests_to_apply',
                                       allowed_values=self.allowed_stats_tests)
        else:
            self.stats_tests_to_apply = []

        self.stats_tests_objects = []
        if len(self.stats_tests_to_apply) > 0:
            warnings.warn("Computing statistics for distributions is an experimental feature. While using it, keep in "
                          "mind that the samples of metrics might be correlated.")
            for test_name in self.stats_tests_to_apply:
                self.stats_tests_objects.append(DistributionStatistics(statistical_test=test_name))

        self.scorers = get_scorers(metrics)

    def fit(self, *args, **kwargs):
        """
        Base fit functionality that should be executed before each fit.
        """

        # Set seed for results reproducibility
        np.random.seed(self.random_state)

        # Initialize the report and results
        self.iterations_results = None
        self.report = None
        self.fitted = True

    def compute(self, metrics=None):
        """
        Reports the statistics.

        Args:
            metrics (str or list of strings, optional):  Name or list of names of metrics to be plotted. If not all
            metrics are presented.

        Returns:
            pd Dataframe: Report that contains the evaluation mean and std on train and test sets for each metric.
        """

        if self.fitted is False:
            raise(NotFittedError('The object has not been fitted. Please run fit() method first'))
        if self.report is None:
            raise(ValueError('Report is None, thus it has not been computed by fit method. Please extend the '
                             'BaseVolatilityEstimator class, overwrite fit method, and within fit run compute_report()'))

        if metrics is None:
            return self.report
        else:
            if not isinstance(metrics, list):
                metrics = [metrics]
            return self.report.loc[metrics]

    def plot(self, metrics=None, bins=10, height_per_subplot=5, width_per_subplot=5):
        """
        Plots distribution of the metric

        Args:
            metrics (str or list of strings, optional):  Name or list of names of metrics to be plotted. If not all
                metrics are presented.

            bins (int, optional):  Number of bins into which histogram is built.

            height_per_subplot (int, optional):  Height of each subplot. Default is 5.

            width_per_subplot (int, optional): Width of each subplot. Default is 5.
        """

        target_report = self.compute(metrics=metrics)

        if target_report.shape[0] >= 1:
            fig, axs = plt.subplots(target_report.shape[0], 2, figsize=(width_per_subplot*2,
                                                                        height_per_subplot*target_report.shape[0]))

            # Enable traversing the axs
            axs = axs.flatten()
            axis_index = 0

            for metric, row in target_report.iterrows():
                train, test, delta = self.get_samples_to_plot(metric_name=metric)

                axs[axis_index].hist(train, alpha=0.5, label='Train {}'.format(metric), bins=bins)
                axs[axis_index].hist(test, alpha=0.5, label='Test {}'.format(metric), bins=bins)
                axs[axis_index].set_title('Distributions {}'.format(metric))
                axs[axis_index].legend(loc='upper right')

                axs[axis_index+1].hist(delta, alpha=0.5, label='Delta {}'.format(metric), bins=bins)
                axs[axis_index+1].set_title('Distributions delta {}'.format(metric))
                axs[axis_index+1].legend(loc='upper right')

                axis_index+=2

            for ax in axs.flat:
                ax.set(xlabel='{} score'.format(metric), ylabel='Results count')

    def get_samples_to_plot(self, metric_name):
        """
        Selects samples to be plotted.

        Args:
            metric_name (str):  Name of metric for which the data should be selected.
        """

        current_metric_results = self.iterations_results[self.iterations_results['metric_name'] == metric_name]
        train = current_metric_results['train_score']
        test = current_metric_results['test_score']
        delta = current_metric_results['delta_score']

        return train, test, delta

    def create_report(self):
        """
        Based on the results for each metric for different sampling, mean and std of distributions of all metrics and
        store them as report.
        """

        unique_metrics = self.iterations_results['metric_name'].unique()

        # Get columns which will be filled
        stats_tests_columns = []
        for stats_tests_object in self.stats_tests_objects:
            stats_tests_columns.append('{} statistic'.format(stats_tests_object.statistical_test_name))
            stats_tests_columns.append('{} p-value'.format(stats_tests_object.statistical_test_name))
        stats_columns = ['train_mean', 'train_std', 'test_mean', 'test_std', 'delta_mean', 'delta_std']
        report_columns = stats_columns + stats_tests_columns

        self.report = pd.DataFrame([], columns=report_columns)

        for metric in unique_metrics:
            metric_iterations_results = self.iterations_results[self.iterations_results['metric_name'] == metric]
            metrics = self.compute_mean_std_from_runs(metric_iterations_results)
            stats_tests_values = self.compute_stats_tests_values(metric_iterations_results)
            metric_row = pd.DataFrame([metrics + stats_tests_values], columns=report_columns, index=[metric])
            self.report = self.report.append(metric_row)

    def compute_mean_std_from_runs(self, metric_iterations_results):
        """
        Compute mean and std of results.

        Args:
            metric_iterations_results (pandas.DataFrame): Scores for a single metric for each iteration.

        Returns:
            list: List containing mean and std of train, test and deltas.
        """
        train_mean_score = np.mean(metric_iterations_results['train_score'])
        test_mean_score = np.mean(metric_iterations_results['test_score'])
        delta_mean_score = np.mean(metric_iterations_results['delta_score'])
        train_std_score = np.std(metric_iterations_results['train_score'])
        test_std_score = np.std(metric_iterations_results['test_score'])
        delta_std_score = np.std(metric_iterations_results['delta_score'])
        return [train_mean_score, train_std_score, test_mean_score, test_std_score, delta_mean_score, delta_std_score]

    def compute_stats_tests_values(self, metric_iterations_results):
        """
        Compute statistics and p-values of specified tests.

        Args:
            metric_iterations_results (pandas.DataFrame):  Scores for a single metric for each iteration.

        Returns:
            list: List containing statistics and p-values of distributions.
        """
        statistics = []
        for stats_test in self.stats_tests_objects:
            stats, p_value = \
                stats_test.compute(metric_iterations_results['test_score'], metric_iterations_results['train_score'])
            statistics += [stats, p_value]
        return statistics

    def fit_compute(self,  *args, **kwargs):
        """
        Runs trains and evaluates a number of models on train and test sets extracted using different random seeds.
        Reports the statistics of the selected metric.

        Returns:
            pandas.Dataframe: Report that contains the evaluation mean and std on train and test sets for each metric.
        """

        self.fit(*args, **kwargs)
        return self.compute()


class TrainTestVolatility(BaseVolatilityEstimator):
    """
    Estimation of volatility of metrics. The estimation is done by splitting the data into train and test multiple times
    and training and scoring a model based on these metrics. The class allows for choosing whether at each iteration
    the train test split should be the same or different, whether and how the train and test sets should be sampled.

    Args:
        model (model object): Binary classification model or pipeline.

        iterations: (int, optional) Number of iterations in seed bootstrapping. By default 1000.

        sample_train_test_split_seed (bool, optional): Flag indicating whether each train test split should be done
        randomly or measurement should be done for single split. Default is True, which indicates that each.
        iteration is performed on a random train test split. If the value is False, the random_seed for the
        split is set to train_test_split_seed.

        train_sampling_type (str): String indicating what type of sampling should be applied on train set:

                    - None indicates that no additional sampling is done after splitting data,

                    - 'bootstrap' indicates that sampling with replacement will be performed on train data,

                    - 'subsample': indicates that sampling without repetition will be performed  on train data.

        test_sampling_type (str, optional): String indicating what type of sampling should be applied on test set:

                    - None indicates that no additional sampling is done after splitting data,

                    - 'bootstrap' indicates that sampling with replacement will be performed on test data,

                    - 'subsample': indicates that sampling without repetition will be performed  on test data.

        train_sampling_fraction: (float, optional): Fraction of train data sampled, if sample_train_type is not None.
        Default value is 1.

        test_sampling_fraction: (Optional float): Fraction of test data sampled, if sample_test_type is not None.
        Default value is 1.

        metrics (string, list of strings, Scorer or list of Scorers): Metrics for which the score is calculated.
        It can be either a name or list of names metric names and needs to be aligned with predefined classification
        scorers names in sklearn, see the `sklearn documentation <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
        In case a custom metric is used, one can create own Scorer (probatus.utils) and provide as a metric.
        By default 'roc_auc' is measured.

        test_prc (float, optional):  Percentage of input data used as test. By default 0.25.

        n_jobs (int, optional): Number of parallel executions. If -1 use all available cores. By default 1.

        stats_tests_to_apply (None, string or list of strings, optional):  List of tests to apply. Available options:

                    - 'ES': Epps-Singleton,

                    - 'KS': Kolmogorov-Smirnov statistic,

                    - 'PSI': Population Stability Index,

                    - 'SW': Shapiro-Wilk based difference statistic,

                    - 'AD': Anderson-Darling TS.

        random_state (int, optional): The seed used by the random number generator.

    Examples:
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from probatus.metric_volatility import TrainTestVolatility
        >>> X, y = make_classification(n_features=4)
        >>> clf = RandomForestClassifier()
        >>> volatility = TrainTestVolatility(clf, iterations=500 , test_prc = 0.5)
        >>> volatility_report = volatility.fit_compute(X, y)
        >>> volatility.plot()
    """

    def __init__(self, model, iterations=1000, sample_train_test_split_seed=True, train_sampling_type=None,
                 test_sampling_type=None, train_sampling_fraction=1, test_sampling_fraction=1, **kwargs):
        super().__init__(model=model, **kwargs)
        self.iterations = iterations
        self.train_sampling_type = train_sampling_type
        self.test_sampling_type = test_sampling_type
        self.sample_train_test_split_seed=sample_train_test_split_seed
        self.train_sampling_fraction = train_sampling_fraction
        self.test_sampling_fraction = test_sampling_fraction

        check_sampling_input(train_sampling_type, train_sampling_fraction, 'train')
        check_sampling_input(test_sampling_type, test_sampling_fraction, 'test')

    def fit(self, X, y):
        """
        Bootstraps a number of random seeds, then splits the data based on the sampled seeds and estimates performance
        of the model based on the split data.

        Args
            X (pandas.DataFrame or numpy.ndarray):  Array with samples and features.

            y (pandas.DataFrame or numpy.ndarray):  Array with targets.

        """
        super().fit()

        X = assure_numpy_array(X)
        y = assure_numpy_array(y)

        if self.sample_train_test_split_seed:
            random_seeds = np.random.random_integers(0, 999999, self.iterations)
        else:
            random_seeds = (np.ones(self.iterations) * self.random_state).astype(int)

        results_per_iteration = Parallel(n_jobs=self.n_jobs)(delayed(get_metric)(
            X=X, y=y, model=self.model, test_size=self.test_prc, split_seed=split_seed,
            scorers=self.scorers, train_sampling_type=self.train_sampling_type,
            test_sampling_type=self.test_sampling_type, train_sampling_fraction=self.train_sampling_fraction,
            test_sampling_fraction=self.test_sampling_fraction
        ) for split_seed in tqdm(random_seeds))

        self.iterations_results = pd.concat(results_per_iteration, ignore_index=True)

        self.create_report()


class SplitSeedVolatility(TrainTestVolatility):
    """
    Estimation of volatility of metrics depending on the seed used to split the data. At every iteration it splits the
    data into train and test set using a different stratified split and volatility of the metrics is calculated.

    Args:
        model (model object): Binary classification model or pipeline.

        iterations: (int, optional) Number of iterations in seed bootstrapping. By default 1000.

        metrics (string, list of strings, Scorer or list of Scorers): Metrics for which the score is calculated.
        It can be either a name or list of names metric names and needs to be aligned with predefined classification
        scorers names in sklearn, see the `sklearn documentation <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
        In case a custom metric is used, one can create own Scorer (probatus.utils) and provide as a metric.
        By default 'roc_auc' is measured.

        test_prc (float, optional):  Percentage of input data used as test. By default 0.25.

        n_jobs (int, optional): Number of parallel executions. If -1 use all available cores. By default 1.

        stats_tests_to_apply (None, string or list of strings, optional):  List of tests to apply. Available options:

                    - 'ES': Epps-Singleton,

                    - 'KS': Kolmogorov-Smirnov statistic,

                    - 'PSI': Population Stability Index,

                    - 'SW': Shapiro-Wilk based difference statistic,

                    - 'AD': Anderson-Darling TS.

        random_state (int, optional): The seed used by the random number generator.

    Examples:
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from probatus.metric_volatility import SplitSeedVolatility
        >>> X, y = make_classification(n_features=4)
        >>> clf = RandomForestClassifier()
        >>> volatility = SplitSeedVolatility(clf, iterations=500 , test_prc = 0.5)
        >>> volatility_report = volatility.fit_compute(X, y)
        >>> volatility.plot()
    """

    def __init__(self, model, **kwargs):
        super().__init__(model=model, sample_train_test_split_seed=True, train_sampling_type=None,
                         test_sampling_type=None, train_sampling_fraction=1,  test_sampling_fraction=1, **kwargs)


class BootstrappedVolatility(TrainTestVolatility):
    """
    Estimation of volatility of metrics by bootstrapping both train and test set. By default at every iteration the
    train test split is the same. The test shows volatility of metric with regards to sampling different rows from
    static train and test sets.

    Args:
        model (model object): Binary classification model or pipeline.

        iterations: (int, optional) Number of iterations in seed bootstrapping. By default 1000.

        train_sampling_fraction: (float, optional): Fraction of train data sampled, if sample_train_type is not None.
        Default value is 1.

        test_sampling_fraction: (Optional float): Fraction of test data sampled, if sample_test_type is not None.
        Default value is 1.

        metrics (string, list of strings, Scorer or list of Scorers): Metrics for which the score is calculated.
        It can be either a name or list of names metric names and needs to be aligned with predefined classification
        scorers names in sklearn, see the `sklearn documentation <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
        In case a custom metric is used, one can create own Scorer (probatus.utils) and provide as a metric.
        By default 'roc_auc' is measured.

        test_prc (float, optional):  Percentage of input data used as test. By default 0.25.

        n_jobs (int, optional): Number of parallel executions. If -1 use all available cores. By default 1.

        stats_tests_to_apply (None, string or list of strings, optional):  List of tests to apply. Available options:

                    - 'ES': Epps-Singleton,

                    - 'KS': Kolmogorov-Smirnov statistic,

                    - 'PSI': Population Stability Index,

                    - 'SW': Shapiro-Wilk based difference statistic,

                    - 'AD': Anderson-Darling TS.

        random_state (int, optional): The seed used by the random number generator.

    Examples:
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from probatus.metric_volatility import BootstrappedVolatility
        >>> X, y = make_classification(n_features=4)
        >>> clf = RandomForestClassifier()
        >>> volatility = BootstrappedVolatility(clf, iterations=500 , test_prc = 0.5)
        >>> volatility_report = volatility.fit_compute(X, y)
        >>> volatility.plot()
    """

    def __init__(self, model, **kwargs):
        super().__init__(model=model, sample_train_test_split_seed=False, train_sampling_type='bootstrap',
                         test_sampling_type='bootstrap', **kwargs)



import matplotlib.pyplot as plt
from matplotlib.pyplot import figure

def plot_distributions_of_feature(feature_distributions, feature_name=None, sample_names=None, plot_bw_method=0.05,
                                  plot_perc_outliers_removed=0.01, plot_figsize=(15, 6)):
    """
    This function plots multiple distributions of the same feature. It is e.g. useful to compare
    distribution between train and test.

    For categorical feature the plot bar is plotted, and for numeric the density plot.

    Args:
        feature_distributions (list of pd.Series): List of distributions of the same feature, e.g. values of feature 'f1'
        for Train, Validation and Test.

        feature_name (Optional, str): Name of the feature plotted.

        sample_names (Optional, list of str): List of names of samples e.g. ['Train', 'Validation', 'Test'].

        plot_bw_method (Optional, float): Estimator bandwidth in density plot.

        plot_perc_outliers_removed (Optional, float): Percentage of outliers removed from each side before plotting.

        plot_figsize (Optional, tuple): Size of the figure.

    """
    figure(figsize=plot_figsize)

    if feature_name is None:
        feature_name = feature_distributions[0].name

    if sample_names is None:
        sample_names = [f'sample_{i}' for i in range(len(feature_distributions))]

    if feature_distributions[0].dtype.name == 'category':
        data_dict = {}

        for feature_distribution_index in range(len(feature_distributions)):
            data_dict[sample_names[feature_distribution_index]] = \
                feature_distributions[feature_distribution_index].value_counts(normalize=True)

        plt.ylabel('Relative frequencies of values in feature.')
    else:
        for feature_distribution_index in range(len(feature_distributions)):
            current_feature = feature_distributions[feature_distribution_index]

            # Remove outliers in each feature
            current_feature = current_feature[current_feature.between(current_feature.quantile(0 + plot_perc_outliers_removed),
                                                                      current_feature.quantile(1 - plot_perc_outliers_removed))]

            # Plot density plot
            current_feature.plot.density(bw_method=plot_bw_method)

    plt.title(f'Distribution of {feature_name}')
    plt.xlabel('Feature Distribution')
    plt.legend(sample_names, loc='upper right')
    plt.show()


import numpy as np
from probatus.utils.tree import TreePathFinder
from probatus.utils.warnings import NotIntendedUseWarning
import pytest


class mock_sub_tree():
    def __init__(self, node_count, cl, cr, feat, thresh):
        self.node_count = node_count
        self.children_left = cl
        self.children_right = cr
        self.feature = feat
        self.threshold = thresh

class mock_tree():
    def __init__(self, node_count, cl, cr, feat, thresh):
        self.tree_ = mock_sub_tree(node_count, cl, cr, feat, thresh)

@pytest.fixture(scope='function')
def make_tree_mock():
    return mock_tree(
        node_count=7,
        cl=np.array([1, -1, 3, 4, -1, -1, -1]),
        cr=np.array([2, -1, 6, 5, -1, -1, -1]),
        feat=np.array([0, -2, 0, 0, -2, -2, -2]),
        thresh=np.array([-3.54, -2., 2.795, -1.985, -2., -2., -2.])
    )

@pytest.fixture(scope='function')
def make_complex_tree_mock():
    return mock_tree(
        node_count=15,
        cl=np.array([1, -1, 3, 4, 5, -1, -1, 8, -1, -1, 11, 12, -1, -1, -1]),
        cr=np.array([2, -1, 10, 7, 6, -1, -1, 9, -1, -1, 14, 13, -1, -1, -1]),
        feat=np.array([0, -2, 0, 0, 0, -2, -2, 0, -2, -2, 0, 0, -2, -2, -2]),
        thresh=np.array([-3.54, -2, 2.795, -1.985, -2.785, -2., -2., 1.675, -2., -2., 3.925, 3.865, -2., -2., -2.])
    )

@pytest.fixture(scope='function')
def make_wrong_tree_mock():
    return mock_tree(
        node_count=7,
        cl=np.array([1, -1, 3, 4, -1, -1, -1]),
        cr=np.array([2, -1, 6, 5, -1, -1, -1]),
        feat=np.array([0, -2, 0, 1, -2, -2, -2]), # this tree is trained on two features, not ideal for binning
        thresh=np.array([-3.54, -2., 2.795, -1.985, -2., -2., -2.])
    )


def test_simple_tree(make_tree_mock):
    # Quick interpretation tip:
    # The sklearn trees structure is represented by the arrays described in the mock tree.
    # When the index is negative, this means that the leaf is a terminal leaf, otherwise is part of the node
    # children left/right contain the index of the next leaf, while feat contains the index of the feature used.
    # In this toy case, the leaf are at index 1,4,5,6 (corresponding to the negative indexes in the arrays).
    # Furthermore, the order of the expected boundaries can be extracted from the thesholds

    # test_tree_1 =

    tpf1 = TreePathFinder(make_tree_mock)


    exp_boundaries = {1: {'min': -np.inf, 'max': -3.54},
                      4: {'min': -3.54, 'max': -1.985},
                      5: {'min': -1.985, 'max': 2.795},
                      6: {'min': 2.795, 'max': np.inf}}


    assert tpf1.get_boundaries() == exp_boundaries
    assert all(tpf1.is_leaves[[1,4,5,6]])#

def test_complex_tree(make_complex_tree_mock):


    tpf2 = TreePathFinder(make_complex_tree_mock)

    exp_boundaries = {1: {'min': -np.inf, 'max': -3.54},
                      5: {'min': -3.54, 'max': -2.785},
                      6: {'min': -2.785, 'max': -1.985},
                      8: {'min': -1.985, 'max': 1.675},
                      9: {'min': 1.675, 'max': 2.795},
                      12: {'min': 2.795, 'max': 3.865},
                      13: {'min': 3.865, 'max': 3.925},
                      14: {'min': 3.925, 'max': np.inf}}

    assert tpf2.get_boundaries() == exp_boundaries
    assert all(tpf2.is_leaves[[1, 5, 6, 8,9,12,13,14]])

def test_warning_raised(make_wrong_tree_mock):
    # This specific tree is trained on two features (feat array has index 0 and 1)
    # The boundaries found by the TreePathFinder are indended for use only on 1 dimensions
    # Hence this tests that a warning is being raised in case this happens

    with pytest.warns(NotIntendedUseWarning):
        TreePathFinder(make_wrong_tree_mock).get_boundaries()




import numpy as np
import pandas as pd
import numbers

from probatus.utils import DimensionalityError
import warnings


def check_1d(x):
    """
    Checks whether or not a list, numpy array, pandas dataframe, pandas series are one-dimensional.
    Returns True when check is ok, otherwise throws a `DimensionalityError`

    Args:
        x: list, numpy array, pandas dataframe, pandas series

    Returns: True or throws `DimensionalityError`

    """
    if isinstance(x, list):
        if any([isinstance(el, list) for el in x]):
            raise DimensionalityError('The input is not 1D')
        else:
            return True
    if isinstance(x, np.ndarray):
        if x.ndim == 1 and all([isinstance(el, numbers.Number) for el in x]):
            return True
        else:
            raise DimensionalityError('The input is not 1D')
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1 and pd.api.types.is_numeric_dtype(
                x[x.columns[0]]):
            return True
        else:
            raise DimensionalityError('The input is not 1D')
    if isinstance(x, pd.core.series.Series):
        if x.ndim == 1 and pd.api.types.is_numeric_dtype(x):
            return True
        else:
            raise DimensionalityError('The input is not 1D')


def assure_numpy_array(x, assure_1d=False):
    """
    Returns x as numpy array. X can be a list, numpy array, pandas dataframe, pandas series

    Args:
        x: list, numpy array, pandas dataframe, pandas series
        assure_1d: whether or not to assure that the input x is one-dimensional

    Returns: numpy array

    """
    if assure_1d:
        _ = check_1d(x)
    if isinstance(x, list):
        return np.array(x)
    if isinstance(x, np.ndarray):
        return x
    if isinstance(x, pd.core.frame.DataFrame):
        if len(x.columns) == 1:
            return x.values.flatten()
        else:
            return x.values
    if isinstance(x, pd.core.series.Series):
        return x.values


def assure_pandas_df(x):
    """
    Returns x as pandas DataFrame. X can be a list, list of lists, numpy array, pandas DataFrame or pandas Series
    
    Args:
        x (list, numpy array, pandas DataFrame, pandas Series): array to be tested
        
    Returns:
        pandas DataFrame
    """
    if isinstance(x, pd.core.frame.DataFrame):
        return x
    elif any([
            isinstance(x, np.ndarray),
            isinstance(x, pd.core.series.Series),
            isinstance(x, list)
    ]):
        return pd.DataFrame(x)
    else:
        raise TypeError(
            "Please supply a list, numpy array, pandas Series or pandas DataFrame"
        )


def warn_if_missing(variable, variable_name):
    """
    Checks if for missing values: if there are notify the user

    Args:
        variable (pandas.DataFrame, pandas.Series or numpy.ndarray): data to be checked for missing values.
        variable_name (str): Name of the variable checked.
    """
    warning_text = "You have missing values in your variable {}, this might cause the model to fail. Please either " \
                   "impute the missing values or use a model that can handle them e.g. XGBoost.".format(variable_name)

    if isinstance(variable, (pd.DataFrame, pd.Series)):
        if variable.isnull().values.any():
            warnings.warn(warning_text)
    if isinstance(variable, np.ndarray):
        if np.isnan(variable).any():
            warnings.warn(warning_text)


def check_numeric_dtypes(x):
    """
    Checks if all entries in an array are of a data type that can be interpreted as numeric (int, float or bool)
    
    Args:
        x (np.ndarray or pd.Series, list): array to be checked
    
    Returns:
        x: unchanged input array
        
    Raises:
        TypeError: if not all elements are of numeric dtypes
    """
    x = assure_numpy_array(x)
    allowed_types = [bool, int, float]
    
    for element in np.nditer(x):
        if type(element.item()) not in allowed_types:
            raise TypeError("Please supply an array with only floats, ints or booleans")
    return x

